@@ -2,33 +2,40 @@
 # encoding: utf-8
 
 __author__ = "ChenyangGao <https://chenyanggao.github.io>"
-__version__ = (0, 0, 2)
-__all__ = ["thread_batch", "thread_pool_batch", "async_batch", "threaded", "run_as_thread"]
+__version__ = (0, 0, 5)
+__all__ = [
+    "thread_batch", "thread_pool_batch", "async_batch", 
+    "threaded", "run_as_thread", "asynchronized", "run_as_async", 
+    "threadpool_map", "taskgroup_map", 
+]
 
-from asyncio import CancelledError, Semaphore as AsyncSemaphore, TaskGroup
-from collections.abc import Callable, Coroutine, Iterable
-from concurrent.futures import Future, ThreadPoolExecutor
+from asyncio import (
+    ensure_future, get_event_loop, BaseEventLoop, CancelledError, 
+    Future as AsyncFuture, Semaphore as AsyncSemaphore, TaskGroup, 
+)
+from collections.abc import (
+    AsyncIterator, Awaitable, Callable, Coroutine, Iterable, 
+    Iterator, Mapping, 
+)
+from concurrent.futures import Executor, Future, ThreadPoolExecutor
 from functools import partial, update_wrapper
-from inspect import isawaitable
+from inspect import isawaitable, iscoroutinefunction
 from os import cpu_count
-from queue import Queue
+from queue import Queue, Empty
+from _thread import start_new_thread
 from threading import Event, Lock, Semaphore, Thread
-from typing import cast, Any, ContextManager, Optional, ParamSpec, TypeVar
+from typing import cast, Any, ContextManager, Literal
 
 from argtools import argcount
+from asynctools import ensure_coroutine
 from decotools import optional
 
 
-Args = ParamSpec("Args")
-T = TypeVar("T")
-V = TypeVar("V")
-
-
-def thread_batch(
+def thread_batch[T, V](
     work: Callable[[T], V] | Callable[[T, Callable], V], 
     tasks: Iterable[T], 
-    callback: Optional[Callable[[V], Any]] = None, 
-    max_workers: Optional[int] = None, 
+    callback: None | Callable[[V], Any] = None, 
+    max_workers: None | int = None, 
 ):
     ac = argcount(work)
     if ac < 1:
@@ -36,13 +43,24 @@ def thread_batch(
     with_submit = ac > 1
     if max_workers is None or max_workers <= 0:
         max_workers = min(32, (cpu_count() or 1) + 4)
-    sentinal = object()
+
     q: Queue[T | object] = Queue()
     get, put, task_done = q.get, q.put, q.task_done
+    sentinal = object()
     lock = Lock()
+    nthreads = 0
+    running = True
+
     def worker():
         task: T | object
-        while (task := get()) is not sentinal:
+        while running:
+            try:
+                task = get(timeout=1)
+            except Empty:
+                continue
+            if task is sentinal:
+                put(sentinal)
+                break
             task = cast(T, task)
             try:
                 if with_submit:
@@ -53,32 +71,33 @@ def thread_batch(
                     callback(r)
             except BaseException:
                 pass
-            task_done()
-        put(sentinal)
-    nthreads = 0
+            finally:
+                task_done()
+
     def submit(task):
         nonlocal nthreads
-        if nthreads == max_workers:
-            put(task)
-        else:
+        if nthreads < max_workers:
             with lock:
                 if nthreads < max_workers:
-                    Thread(target=worker, daemon=True).start()
+                    start_new_thread(worker, ())
                     nthreads += 1
+        put(task)
+
     for task in tasks:
         submit(task)
     try:
         q.join()
     finally:
+        running = False
         q.queue.clear()
         put(sentinal)
 
 
-def thread_pool_batch(
+def thread_pool_batch[T, V](
     work: Callable[[T], V] | Callable[[T, Callable], V], 
     tasks: Iterable[T], 
-    callback: Optional[Callable[[V], Any]] = None, 
-    max_workers: Optional[int] = None, 
+    callback: None | Callable[[V], Any] = None, 
+    max_workers: None | int = None, 
 ):
     ac = argcount(work)
     if ac < 1:
@@ -86,9 +105,11 @@ def thread_pool_batch(
     with_submit = ac > 1
     if max_workers is None or max_workers <= 0:
         max_workers = min(32, (cpu_count() or 1) + 4)
+
     ntasks = 0
     lock = Lock()
     done_evt = Event()
+
     def works(task):
         nonlocal ntasks
         try:
@@ -103,11 +124,13 @@ def thread_pool_batch(
                 ntasks -= 1
             if not ntasks:
                 done_evt.set()
+
     def submit(task):
         nonlocal ntasks
         with lock:
            ntasks += 1
         return create_task(works, task)
+
     pool = ThreadPoolExecutor(max_workers)
     try:
         create_task = pool.submit
@@ -119,11 +142,11 @@ def thread_pool_batch(
         pool.shutdown(False, cancel_futures=True)
 
 
-async def async_batch(
+async def async_batch[T, V](
     work: Callable[[T], Coroutine[None, None, V]] | Callable[[T, Callable], Coroutine[None, None, V]], 
     tasks: Iterable[T], 
-    callback: Optional[Callable[[V], Any]] = None, 
-    sema: Optional[AsyncSemaphore] = None, 
+    callback: None | Callable[[V], Any] = None, 
+    sema: None | AsyncSemaphore = None, 
 ):
     ac = argcount(work)
     if ac < 1:
@@ -159,15 +182,15 @@ async def async_batch(
 
 
 @optional
-def threaded(
-    func: Callable[Args, Any], 
+def threaded[**Args, T](
+    func: Callable[Args, T], 
     /, 
     lock: None | int | ContextManager = None, 
     **thread_init_kwds, 
-) -> Callable[Args, Future]:
+) -> Callable[Args, Future[T]]:
     if isinstance(lock, int):
         lock = Semaphore(lock)
-    def wrapper(*args: Args.args, **kwds: Args.kwargs) -> Future[V]:
+    def wrapper(*args: Args.args, **kwds: Args.kwargs) -> Future[T]:
         if lock is None:
             def asfuture():
                 try: 
@@ -181,14 +204,14 @@ def threaded(
                         fu.set_result(func(*args, **kwds))
                     except BaseException as e:
                         fu.set_exception(e)
-        fu: Future[V] = Future()
+        fu: Future[T] = Future()
         thread = fu.thread = Thread(target=asfuture, **thread_init_kwds) # type: ignore
         thread.start()
         return fu
     return update_wrapper(wrapper, func)
 
 
-def run_as_thread(
+def run_as_thread[**Args, T](
     func: Callable[Args, T], 
     /, 
     *args: Args.args, 
@@ -196,3 +219,78 @@ def run_as_thread(
 ) -> Future[T]:
     return getattr(threaded, "__wrapped__")(func)(*args, **kwargs)
 
+
+@optional
+def asynchronized[**Args, T](
+    func: Callable[Args, T], 
+    /, 
+    loop: None | BaseEventLoop = None, 
+    executor: Literal[False, None] | Executor = None, 
+) -> Callable[Args, AsyncFuture[T]]:
+    def run_in_thread(loop, future, args, kwargs, /):
+        try:
+            loop.call_soon_threadsafe(future.set_result, func(*args, **kwargs))
+        except BaseException as e:
+            loop.call_soon_threadsafe(future.set_exception, e)
+    def wrapper(*args: Args.args, **kwargs: Args.kwargs) -> AsyncFuture[T]:
+        nonlocal loop
+        if iscoroutinefunction(func):
+            return ensure_future(func(*args, **kwargs), loop=loop)
+        if loop is None:
+            loop = cast(BaseEventLoop, get_event_loop())
+        if executor is False:
+            future = loop.create_future()
+            start_new_thread(run_in_thread, (loop, future, args, kwargs))
+            return future
+        return loop.run_in_executor(executor, partial(func, *args, **kwargs))
+    return wrapper
+
+
+def run_as_async[**Args, T](
+    func: Callable[Args, T], 
+    /, 
+    *args: Args.args, 
+    **kwargs: Args.kwargs, 
+) -> AsyncFuture[T]:
+    return getattr(asynchronized, "__wrapped__")(func)(*args, **kwargs)
+
+
+def threadpool_map[T](
+    func: Callable[..., T], 
+    it, 
+    /, 
+    *its, 
+    max_workers: None | int = None, 
+    kwargs: None | Mapping = None, 
+) -> Iterator[T]:
+    if kwargs:
+        func = partial(func, **kwargs)
+    executor = ThreadPoolExecutor(max_workers=max_workers)
+    try:
+        with executor:
+            yield from executor.map(func, it, *its)
+    finally:
+        executor.shutdown(wait=False, cancel_futures=True)
+
+
+async def taskgroup_map[T](
+    func: Callable[..., Awaitable[T]], 
+    it, 
+    /, 
+    *its, 
+    max_workers: None | int = None, 
+    kwargs: None | Mapping = None, 
+) -> AsyncIterator[T]:
+    if max_workers is None:
+        max_workers = 20
+    if kwargs is None:
+        kwargs = {}
+    sema = AsyncSemaphore(max_workers)
+    async def call(args):
+        async with sema:
+            return await func(*args, **kwargs)
+    async with TaskGroup() as tg:
+        create_task = tg.create_task
+        for task in [create_task(call(args)) for args in zip(it, *its)]:
+            yield await task
+
